
namespace Midori
{
	public enum SecurityStatus
	{
		/**
		 * signals that the current view has no specified security status. It may
		 * be a native view.
		 */
		UNSPECIFIED,
		/**
		 * signals that the current view has no secure connection
		 */
		NONE,
		/**
		 * signals that the current view has an unknown secure connection
		 */
		UNKNOWN,
		/**
		 * signals that the current view has a known secure connection
		 */
		TRUSTED
	}

	public class Tab : Gtk.Stack, Searchable, Navigatable, Notifier, Savable, Zoomable, ResourceContainer
	{
		TabContent _current_content;
		public TabContent current_content {
			get {
				return _current_content;
			}
			set {
				if (value == _current_content)
					return;

				// disconnect from previous instance
				if (_current_content != null) {
					if (_current_content is Navigatable) {
						var navigatable = (Navigatable) _current_content;
						navigatable.request_open_new_window.disconnect (_request_open_new_window);
						navigatable.add_history_entry.disconnect (_add_history_entry);
					}

					if (_current_content is Notifier) {
						var notifier = (Notifier) _current_content;
						notifier.open_dialog.disconnect (_open_dialog);
					}

					if (_current_content is ResourceContainer) {
						var resource_container = (ResourceContainer) _current_content;
						resource_container.download_resource.disconnect (_download_resource);
					}

					if (_current_content is Zoomable) {
						var zoomable = (Zoomable) _current_content;
						zoomable.notify["zoom-level"].disconnect (_notify_zoom_level);
					}

					if (_current_content is Loadable) {
						var loadable = (Loadable) _current_content;
						loadable.notify["loading"].disconnect (_notify_loading);
						loadable.notify["estimated-load-progress"].disconnect (_notify_progress);
					}

					_current_content.notify["page-title"].disconnect (_notify_page_title);
					_current_content.notify["icon"].disconnect (_notify_icon);
					_current_content.notify["ui-color"].disconnect (_notify_ui_color);
				}

				_current_content = value;

				// connect to new instance
				if (_current_content is Navigatable) {
					var navigatable = (Navigatable) _current_content;
					navigatable.request_open_new_window.connect (_request_open_new_window);
					navigatable.add_history_entry.connect (_add_history_entry);
				}

				if (_current_content is Notifier) {
					var notifier = (Notifier) _current_content;
					notifier.open_dialog.connect (_open_dialog);
				}

				if (_current_content is ResourceContainer) {
					var resource_container = (ResourceContainer) _current_content;
					resource_container.download_resource.disconnect (_download_resource);
				}

				if (_current_content is Zoomable) {
					var zoomable = (Zoomable) _current_content;
					zoomable.notify["zoom-level"].connect (_notify_zoom_level);
				}

				if (_current_content is Loadable) {
					var loadable = (Loadable) _current_content;
					loadable.notify["loading"].connect (_notify_loading);
					loadable.notify["estimated-load-progress"].connect (_notify_progress);
				}

				_current_content.notify["page-title"].connect (_notify_page_title);
				_current_content.notify["icon"].connect (_notify_icon);
				_current_content.notify["ui-color"].connect (_notify_ui_color);
			}
		}

		/**
		 * The uri of the currently loaded page
		 */
		public string uri { get; private set; default = "about:blank"; }

		/**
		 * @inheritDoc
		 */
		public bool loading {
			get {
				var loadable = current_content as Loadable;
				if (loadable != null)
					return loadable.loading;
				else
					return false;
			}
			protected set {
			}
		}
		void _notify_loading (Object object, ParamSpec pspec)
		{
			notify_property ("loading");
		}

		/**
		 * @inheritDoc
		 */
		public double progress {
			get {
				var loadable = current_content as Loadable;
				if (loadable != null)
					return loadable.progress;
				else
					assert_not_reached ();
			}
			protected set {
			}
		}
		void _notify_progress (Object object, ParamSpec pspec)
		{
			notify_property ("progress");
		}

		/**
		 * @inheritDoc
		 */
		public double zoom_level {
			get {
				var loadable = current_content as Loadable;
				if (loadable != null)
					return loadable.progress;
				else
					assert_not_reached ();
			}
			set {
				var zoomable = current_content as Zoomable;

				if (zoomable != null)
					zoomable.zoom_level = value;
				else
					assert_not_reached ();
			}
		}
		void _notify_zoom_level (Object object, ParamSpec pspec)
		{
			notify_property ("zoom-level");
		}

		public string page_title {
			get {
				return _current_content.page_title;
			}
			protected set {
			}
		}
		void _notify_page_title (Object object, ParamSpec pspec)
		{
			notify_property ("page-title");
		}

		public Gdk.Pixbuf? icon {
			get {
				return _current_content.icon;
			}
			protected set
			{
			}
		}
		void _notify_icon (Object object, ParamSpec pspec)
		{
			notify_property ("icon");
		}

		public string? ui_color {
			get {
				return _current_content.ui_color;
			}
		}
		void _notify_ui_color (Object object, ParamSpec pspec)
		{
			notify_property ("ui-color");
		}

		public bool can_go_back {
			get {
				return web_view.can_go_back ();
			}
		}

		public bool can_go_forward {
			get {
				return web_view.can_go_forward ();
			}
		}

		public SecurityStatus security_status { get; private set; default = SecurityStatus.UNSPECIFIED; }

		/**
		 * Each Tab has a web_view that will for example be moved to
		 * the foreground when a page load is requested via urlbar
		 */
		WebView web_view;
		Gtk.ScrolledWindow web_view_scrolled;

		public Tab ()
		{
			web_view = new WebView ();
			web_view_scrolled = new Gtk.ScrolledWindow (null, null);
			web_view_scrolled.kinetic_scrolling = true;

			web_view.get_back_forward_list ().changed.connect ((added, removed) => {
				notify_property ("can-go-forward");
				notify_property ("can-go-back");

				_debug_navigation_stack ();
			});
			web_view.notify["uri"].connect (() => {
				if (web_view.uri.has_prefix ("about:"))
					visible_child_name = web_view.uri.split (":")[1];
				else if (visible_child_name != "web-view")
					visible_child_name = "web-view";

				uri = web_view.uri;

				_debug_navigation_stack ();
			});
			web_view.load_changed.connect (web_view_load_changed);

			current_content = web_view;

			web_view_scrolled.add (web_view);

			add_named (web_view_scrolled, "web-view");
		}

		public override void grab_focus ()
		{
			current_content.grab_focus ();
		}

		public bool implements (Type iface)
		{
			return current_content.get_type ().is_a (iface);
		}

		/**
		 * Show the webview and load the given uri
		 *
		 * @param uri The uri to be loaded
		 */
		public void go_to_uri (string uri)
		{
			visible_child_name = "web-view";
			web_view.load_uri (uri);
		}

		/**
		 * Show the given native widget
		 *
		 * @param custom_id A custom id to later identify the widget in the history stack
		 * @param view		The native widget to be shown
		 */
		public void go_to_native (string custom_id, Gtk.Widget view)
		{
			add_named (view, custom_id);
			view.show ();

			go_to_uri ("about:" + custom_id);
		}

		/**
		 * @inheritDoc
		 */
		public void search (string text, bool case_insensitive, bool highlight)
		{
			var search = current_content as Searchable;

			if (search != null)
				search.search (text, case_insensitive, highlight);
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 */
		public void next_result ()
		{
			var search = current_content as Searchable;

			if (search != null)
				search.next_result ();
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 */
		public void previous_result ()
		{
			var search = current_content as Searchable;

			if (search != null)
				search.previous_result ();
			else
				assert_not_reached ();
		}

		void _add_history_entry (string custom_id, string? title)
		{
			web_view.load_uri ("about:" + custom_id + ":" + title);

			add_history_entry (custom_id, title);
		}

		/**
		 * @inheritDoc
		 */
		public void open_history_item (string id)
		{
			var navigatable = current_content as Navigatable;

			if (navigatable != null)
				navigatable.open_history_item (id);
			else
				assert_not_reached ();
		}

		void _request_open_new_window (string uri)
		{
			request_open_new_window (uri);
		}

		void _open_dialog (Notifier.DialogType type, string message)
		{
			open_dialog (type, message);
		}

		/**
		 * @inheritDoc
		 */
		public void reload (bool bypass_cache = false)
		{
			var loadable = current_content as Loadable;

			if (loadable != null)
				loadable.reload ();
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 */
		public void abort_loading ()
		{
			var loadable = current_content as Loadable;

			if (loadable != null)
				loadable.abort_loading ();
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 */
		public void zoom_in ()
		{
			var zoomable = current_content as Zoomable;

			if (zoomable != null)
				zoomable.zoom_in ();
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 */
		public void zoom_out ()
		{
			var zoomable = current_content as Zoomable;

			if (zoomable != null)
				zoomable.zoom_out ();
			else
				assert_not_reached ();
		}

		/**
		 * @inheritDoc
		 *
		 * This function won't fail if the interfaces don't match, but return false instead.
		 */
		public bool can_zoom_in ()
		{
			var zoomable = current_content as Zoomable;

			if (zoomable != null)
				return zoomable.can_zoom_in ();
			else
				return false;
		}

		/**
		 * @inheritDoc
		 *
		 * This function won't fail if the interfaces don't match, but return false instead.
		 */
		public bool can_zoom_out ()
		{
			var zoomable = current_content as Zoomable;

			if (zoomable != null)
				return zoomable.can_zoom_out ();
			else
				return false;
		}

		/**
		 * @inheritDoc
		 */
		public void reset_zoom ()
		{
			var zoomable = current_content as Zoomable;

			if (zoomable != null)
				zoomable.reset_zoom ();
			else
				assert_not_reached ();
		}


		void _download_resource (string uri)
		{
			download_resource (uri);
		}

		public void forward ()
		{
			if (!can_go_forward)
				return;

			web_view.go_forward ();
			_debug_navigation_stack ();
		}

		public void back ()
		{
			if (!can_go_back)
				return;

			web_view.go_back ();
			_debug_navigation_stack ();
		}

		void web_view_load_changed (WebKit.LoadEvent event)
		{
			switch (event) {
				case WebKit.LoadEvent.COMMITTED:
					if (!web_view.uri.has_prefix ("https")) {
						security_status = web_view.uri.has_prefix ("about:") ?
							SecurityStatus.UNSPECIFIED : SecurityStatus.NONE;
						break;
					}

					security_status = SecurityStatus.TRUSTED;
#if false
					// TODO we crash when loading an non-http web page with this enabled :(
					//      probably at get_tls_info, but webkit dbg symbols are 700MB O_O
					TlsCertificate tls_cert;
					TlsCertificateFlags tls_flags;

					if (web_view.get_tls_info (out tls_cert, out tls_flags)) {
						security_status = SecurityStatus.TRUSTED;
					} else if (tls_cert != null) {
						var gcr_cert = new Gcr.SimpleCertificate (tls_cert.certificate.data);
						var hostname = URI.parse_hostname (web_view.uri, null);

						try {
							if (Gcr.Trust.is_certificate_pinned (gcr_cert, Gcr.Trust.PURPOSE_SERVER_AUTH, hostname, null)) {
								security_status = SecurityStatus.TRUSTED;
							} else {
								security_status = SecurityStatus.UNKNOWN;

								web_view.stop_loading ();
								/* TODO midori_view_display_error (view, NULL, NULL, NULL, _("Security unknown"),
										midori_location_action_tls_flags_to_string (tls_flags), NULL,
										_("Trust this website"),
										NULL);*/
							}
						} catch (Error e) {
							warning (e.message);
						}
					} else {
						security_status = SecurityStatus.UNKNOWN;
					}
#endif
					break;
			}
		}

		void _debug_navigation_stack ()
		{
			print ("Current History Stack:\n");

			var list = web_view.get_back_forward_list ();

			foreach (var item in list.get_back_list ()) {
				print ("-  %s\n", item.get_uri ());
			}

			if (list.get_current_item () != null)
				print ("=> %s\n", list.get_current_item ().get_uri ());

			foreach (var item in list.get_forward_list ()) {
				print ("+  %s\n", item.get_uri ());
			}

			print ("\n");
		}
	}
}

