
namespace Midori
{
	public class UrlBarCompletion : Gtk.Window
	{
		const int MAX_RESULTS = 10;
		const int DELAY_ADDING_RESULTS = 150;

		public signal void uri_selected (string uri);

		public Gtk.Entry entry { get; construct; }
		public bool working { get; private set; default = false; }

		bool is_grabbing = false;
		Gdk.Device? device = null;

		Gtk.TreeView view;
		Gtk.ListStore list;
		Gtk.ScrolledWindow scroll;
		Icon default_icon;

		Autocompleter completer;

		public UrlBarCompletion (Gtk.Entry entry)
		{
			Object (entry: entry,
					resizable: false,
					type_hint: Gdk.WindowTypeHint.COMBO,
					type: Gtk.WindowType.POPUP);
		}

		construct
		{
			var frame = new Gtk.Frame (null);
			frame.shadow_type = Gtk.ShadowType.ETCHED_IN;

			scroll = new Gtk.ScrolledWindow (null, null);
			scroll.hscrollbar_policy = Gtk.PolicyType.NEVER;

			view = new Gtk.TreeView ();
			view.headers_visible = false;
			view.show_expanders = false;
			view.level_indentation = 12;

			default_icon = new ThemedIcon ("empty");

			get_style_context ().add_class ("completion-popup");

			var column = new Gtk.TreeViewColumn ();
			column.sizing = Gtk.TreeViewColumnSizing.FIXED;

			var cell_icon = new Gtk.CellRendererPixbuf ();
			column.pack_start (cell_icon, false);
			column.set_cell_data_func (cell_icon, (cell_layout, renderer, model, iter) => {
				Icon icon;
				model.get (iter, Autocompleter.Columns.ICON, out icon);

				((Gtk.CellRendererPixbuf) renderer).gicon = icon ?? default_icon;
			});

			var cell_title = new Gtk.CellRendererText ();
			cell_title.ellipsize = Pango.EllipsizeMode.MIDDLE;
			column.pack_start (cell_title, true);
			column.set_cell_data_func (cell_title, (cell_layout, renderer, model, iter) => {
				string text;
				model.get (iter, Autocompleter.Columns.MARKUP, out text);

				((Gtk.CellRendererText) renderer).markup = text.split ("\n")[0];
			});

			var cell_details = new Gtk.CellRendererText ();
			cell_details.xpad = 6;
			cell_details.ellipsize = Pango.EllipsizeMode.MIDDLE;
			column.pack_start (cell_details, false);
			column.set_cell_data_func (cell_details, (cell_layout, renderer, model, iter) => {
				string text;
				model.get (iter, Autocompleter.Columns.MARKUP, out text);

				var parts = text.split ("\n");

				if (parts.length > 1)
					text = parts[1];
				else
					model.get (iter, Autocompleter.Columns.URI, out text);

				((Gtk.CellRendererText) renderer).markup = "<span %s>%s</span>".printf (get_pango_grey_color_string (), text);
			});

			view.append_column (column);

			completer = new Autocompleter ();
			completer.populated.connect (() => {
				Idle.add (() => {
					resize_popup ();
					return false;
				});
			});
			completer.add (new SearchCompletion ());
			// completer.add (new HistoryCompletion ());

			view.model = list = completer.model;

			scroll.add (view);
			frame.add (scroll);
			add (frame);

			entry.focus_out_event.connect (() => {
				popdown ();
				return false;
			});

			button_press_event.connect ((e) => {
				if (e.x >= 0 && e.y >= 0 && e.x < get_allocated_width () && e.y < get_allocated_height ()) {
					view.event (e);
					return false;
				}

				entry.text = "";
				popdown ();
				return false;
			});

			view.button_press_event.connect ((e) => {
				Gtk.TreePath path;
				Gtk.TreeIter iter;

				view.get_path_at_pos ((int) e.x, (int) e.y, out path, null, null, null);

				if (path != null) {
					list.get_iter (out iter, path);
					accept (iter);
				}

				return true;
			});

			key_release_event.connect (key_event);
			key_press_event.connect (key_event);

			entry.key_press_event.connect (entry_key_press);
		}

		bool entry_key_press (Gdk.EventKey event)
		{
			if (!get_mapped ())
				return false;

			switch (event.keyval) {
				case Gdk.Key.Escape:
					popdown ();
					return true;
				case Gdk.Key.Return:
				case Gdk.Key.KP_Enter:
				case Gdk.Key.ISO_Enter:
					accept ();
					return true;
				case Gdk.Key.Up:
				case Gdk.Key.Down:
					if (list_empty ()) {
						Gdk.beep ();
						return true;
					}

					var up = event.keyval == Gdk.Key.Up;

					if (view.get_selection ().count_selected_rows () < 1) {
						if (up)
							select_last ();
						else
							select_first ();
						return true;
					}

					select_adjacent (up);

					return true;
			}

			return false;
		}

		bool key_event (Gdk.EventKey event)
		{
			if (!get_mapped ())
				return false;

			entry.event (event);

			return true;
		}

		void select_first ()
		{
			Gtk.TreeIter iter;
			list.get_iter_first (out iter);

			select_iter (iter);
		}

		void select_last ()
		{
			Gtk.TreeIter iter;
			list.iter_nth_child (out iter, null, list.iter_n_children (null) - 1);

			select_iter (iter);
		}

		void select_adjacent (bool up)
		{
			Gtk.TreeIter iter;
			get_iter_at_cursor (out iter);

			var valid = up ? list.iter_previous (ref iter) : list.iter_next (ref iter);

			if (valid) {
				select_iter (iter);
				return;
			}

			get_iter_at_cursor (out iter);

			if (up)
				select_last ();
			else
				select_first ();
		}

		bool list_empty ()
		{
			return list.iter_n_children (null) < 1;
		}

		void resize_popup ()
		{
			var entry_window = entry.get_window ();
			if (entry_window == null)
				return;

			int x, y;
			Gtk.Allocation entry_alloc;

			entry_window.get_origin (out x, out y);
			entry.get_allocation (out entry_alloc);

			x += entry_alloc.x;
			y += entry_alloc.y;

			var screen = entry.get_screen ();
			var monitor = screen.get_monitor_at_window (entry_window);
			var workarea = screen.get_monitor_workarea (monitor);

			int cell_height, separator_height, items;
			view.style_get ("vertical-separator", out separator_height);
			view.get_column (0).cell_get_size (null, null, null, null, out cell_height);
			items = list.iter_n_children (null);

			var height = items * (cell_height + separator_height) + separator_height;

			if (x < workarea.x)
				x = workarea.x;
			else if (x + width_request > workarea.x + workarea.width)
				x = workarea.x + workarea.width - width_request;

			y += entry_alloc.height;

			if (y + height > workarea.x + workarea.height)
				height = workarea.y + workarea.height - y - 12;

			scroll.set_min_content_height (height);
			set_size_request (int.min (entry_alloc.width, workarea.width), height);
			move (x, y);
			resize (width_request, height_request);
		}

		bool get_iter_at_cursor (out Gtk.TreeIter iter)
		{
			Gtk.TreePath? path = null;

			view.get_cursor (out path, null);
			if (path == null)
				return false;

			list.get_iter (out iter, path);
			return true;
		}

		void select_iter (Gtk.TreeIter iter)
		{
			var path = list.get_path (iter);
			view.set_cursor (path, null, false);
		}

		void popup ()
		{
			if (get_mapped ()
				|| !entry.get_mapped ()
				|| !entry.has_focus
				|| is_grabbing)
				return;

			resize_popup ();

			var toplevel = entry.get_toplevel ();
			if (toplevel is Gtk.Window)
				((Gtk.Window) toplevel).get_group ().add_window (this);

			grab_focus ();
			set_screen (entry.get_screen ());
			show_all ();

			if (device != null) {
				Gtk.device_grab_add (this, device, true);
				device.grab (get_window (), Gdk.GrabOwnership.WINDOW, false, Gdk.EventMask.BUTTON_PRESS_MASK
						| Gdk.EventMask.BUTTON_RELEASE_MASK
						| Gdk.EventMask.POINTER_MOTION_MASK,
						null, Gdk.CURRENT_TIME);

				is_grabbing = true;
			}
		}

		void popdown ()
		{
			entry.reset_im_context ();

			if (is_grabbing && device != null) {
				device.ungrab (Gdk.CURRENT_TIME);
				Gtk.device_grab_remove (this, device);

				is_grabbing = false;
			}

			hide ();
		}

		void accept (Gtk.TreeIter? accepted = null)
		{
			if (list_empty ()) {
				Gdk.beep ();
				return;
			}

			if (accepted == null && !get_iter_at_cursor (out accepted)) {
				uri_selected (entry.text);
				return;
			}

			string? uri = null;
			list.@get (accepted, Autocompleter.Columns.URI, out uri);

			if (uri == null) {
				Gdk.beep ();
				return;
			}

			uri_selected (uri);

			popdown ();
		}

		public void clear ()
		{
			list.clear ();

			resize_popup ();
		}

		public async void search (string term)
		{
			device = Gtk.get_current_event_device ();

			if (device != null && device.input_source == Gdk.InputSource.KEYBOARD)
				device = device.associated_device;

			popup ();

			clear ();
			yield completer.complete (term);
		}

		string get_pango_grey_color_string ()
		{
			Gdk.RGBA rgba;
			string color = "";
			var colored = get_style_context ().lookup_color ("placeholder_text_color", out rgba);

			if (colored) {
				Gdk.Color gdk_color = { 0, (uint16) (rgba.red * 65536), (uint16) (rgba.green * 65536), (uint16) (rgba.blue * 65536) };
				color = "color='%s'".printf (gdk_color.to_string ());
			}

			return color;
		}
	}
}

