
namespace GLib.Hostname
{
	[CCode (cname = "g_hostname_is_ip_address")]
	public extern bool is_ip_address (string hostname);
}

namespace Midori.Utils
{
	public AppInfo? default_for_uri (string uri, out string? scheme)
	{
		scheme = Uri.parse_scheme (uri);
		if (scheme == null)
			return null;

		var info = AppInfo.get_default_for_uri_scheme (scheme);

		return info;
	}

	public bool external_uri (string uri)
	{
		/* URI schemes are case-insensitive, followed by ':' - rfc3986 */
		if (uri == null)
			return false;

		var uri_lower = uri.down ();
		if (uri_lower.has_prefix ("http:")
			|| uri_lower.has_prefix ("https:")
			|| uri_lower.has_prefix ("file:")
			|| uri_lower.has_prefix ("geo:")
			|| uri_lower.has_prefix ("about:"))
			return false;

		var info = default_for_uri (uri, null);

		return info != null;
	}

	/**
	 * Takes a string that was typed by a user,
	 * guesses what it is, and returns an URI.
	 *
	 * If it was a search, %NULL will be returned.
	 *
	 * @param uri a string typed by a user
	 *
	 * @return the URI, or %NULL
	 */
	public string? magic_uri (string? uri, bool allow_search, bool allow_relative)
	{
		if (uri == null)
			return null;

		// Add file:// if we have a local path
		if (Path.is_absolute (uri)) {
			try {
				return Filename.to_uri (uri, null);
			} catch (Error e) {
				warning (e.message);
				return null;
			}
		}

		if (allow_relative && FileUtils.test (uri, FileTest.EXISTS | FileTest.IS_REGULAR)) {
			return File.new_for_commandline_arg (uri).get_uri ();
		}

		// Parse geo URI geo:48.202778,16.368472;crs=wgs84;u=40 as a location
		/* TODO if (uri.has_prefix ("geo:")) {
			gchar* comma;
			gchar* semicolon;
			gchar* latitude;
			gchar* longitude;
			gchar* geo;

			comma = strchr (&uri[4], ',');
			// geo:latitude,longitude[,altitude][;u=u][;crs=crs]
			if (!(comma && *comma))
				return g_strdup (uri);
			semicolon = strchr (comma + 1, ';');
			if (!semicolon)
				semicolon = strchr (comma + 1, ',');
			latitude = g_strndup (&uri[4], comma - &uri[4]);
			if (semicolon)
				longitude = g_strndup (comma + 1, semicolon - comma - 1);
			else
				longitude = g_strdup (comma + 1);
			geo = g_strdup_printf ("http://www.openstreetmap.org/?mlat=%s&mlon=%s",
				latitude, longitude);
			g_free (latitude);
			g_free (longitude);
			return geo;
		}*/

		if (Midori.URI.is_location (uri) || external_uri (uri))
			return uri;

		if (Midori.URI.is_ip_address (uri))
			return "http://" + uri;

		int search = -1;

		if (!(" " in uri) &&
			(((search = uri.index_of_char (':')) >= 0 && uri[search + 1].isdigit ()) || (search = uri.index_of_char ('@')) >= 0))
			return "http://" + uri;

		if ((uri.has_prefix ("localhost") || "/" in uri)
			&& resolve_hostname_sync (uri))
			return "http://" + uri;

		if (search < 0) {
			var parts = uri.split (".");

			if (parts.length > 1 && !(parts[1][1] == '\0' && !(parts[1][0].isalpha ())) &&
				!(" "in parts[0]) && !(" " in parts[1]))
				   return "http://" + uri;
		}

		// FIXME do we want the browser to exit here?
		if (!allow_search)
			error (_("Invalid URI"));

		return null;
	}

	/**
	 * Attempts to prefetch the specified URI, that is
	 * it tries to resolve the hostname in advance.
	 *
	 * @param settings	a #MidoriWebSettings instance, or %NULL
	 * @param uri		an URI string
	 *
	 * @return %TRUE on success
	 */
	public async bool prefetch_uri (WebKit.Settings? settings, string uri)
	{
		if (settings != null && !settings.enable_dns_prefetching)
			return false;

		var hostname = Midori.URI.parse_hostname (uri, null);

		if (hostname == null || Hostname.is_ip_address (hostname) || !Midori.URI.is_http (uri)) {
			return false;
		}

		var address = new Soup.Address (hostname, 80);
		var resolved = false;

		address.resolve_async (null, null, (addr, status) => {
			if (status == Soup.Status.OK)
				resolved = true;

			prefetch_uri.callback ();
		});
		yield;

		// FIXME do we need this if we use soup?
		// var context = WebKit.WebContext.get_default ();
		// context.prefetch_dns (hostname);

		return resolved;
	}

	/**
	 * Takes a string that was typed by a user,
	 * resolves the hostname, and returns the status.
	 *
	 * @param hostname a string typed by a user
	 *
	 * @return %TRUE if is a valid host, else %FALSE
	 */
	public async bool resolve_hostname (string hostname)
	{
		var uri = "http://" + hostname;

		return yield prefetch_uri (null, uri);
	}

	public bool resolve_hostname_sync (string hostname)
	{
		var resolved = false;

		resolve_hostname.begin (hostname, (obj, res) => {
			resolved = resolve_hostname.end (res);
		});

        var timer = new Timer ();
        while (!resolved && timer.elapsed (null) < 10)
			MainContext.@default ().iteration (false);

		print ("RESOLVED: %i\n", (int) resolved);
		return resolved;
	}
}
