
namespace Midori
{
	public class WebView : WebKit.WebView, TabContent, Searchable, Navigatable, Notifier, Savable, Loadable, Zoomable, ResourceContainer
	{
		static Gdk.Pixbuf? fallback_icon;

		static construct
		{
			try {
				fallback_icon = Gtk.IconTheme.get_default ().load_icon ("empty", 16, 0);
			} catch (Error e) {
				fallback_icon = null;
			}
		}

		/**
		 * @inheritDoc
		 */
		public double progress {
			get {
				return 0.0;
			}
			set {
			}
		}

		/**
		 * @inheritDoc
		 */
		public bool loading {
			get {
				return is_loading;
			}
			protected set {
			}
		}

		/**
		 * @inheritDoc
		 */
		public Gdk.Pixbuf? icon {
			get {
				return _icon;
			}
			set {
				_icon = value;
			}
		}
		Gdk.Pixbuf? _icon = fallback_icon;

		/**
		 * @inheritDoc
		 */
		public string page_title {
			get {
				return title ?? uri ?? _("Blank Page");
			}
			protected set {
			}
		}

		/**
		 * @inheritDoc
		 */
		public string? ui_color {
			get {
				return _ui_color;
			}
			protected set {
				_ui_color = value;
			}
		}
		string? _ui_color = null;

		public WebView ()
		{
			notify["is-loading"].connect (() => {
				notify_property ("loading");
			});

			notify["title"].connect (() => {
				notify_property ("page-title");
			});

			notify["favicon"].connect (() => {
				var surface = (Cairo.ImageSurface) get_favicon ();

				if (surface != null) {
					var pixbuf = Gdk.pixbuf_get_from_surface (surface, 0, 0, surface.get_width (), surface.get_height ());
					icon = (pixbuf.get_width () > 16 || pixbuf.get_height () > 16) ?
						pixbuf.scale_simple (16, 16, Gdk.InterpType.HYPER) : pixbuf;
				} else {
					icon = fallback_icon;
				}
			});

			load_changed.connect ((event) => {
				switch (event) {
					case WebKit.LoadEvent.FINISHED:
						determine_theme_color.begin ();
						break;
				}
			});
		}

		/**
		 * Check for the theme-color meta tag in the page and if that one can't be
		 * found grabs the color from the current page and uses the first 3 rows
		 * of pixels to get a good representative color of the page
		 */
		async void determine_theme_color ()
		{
			/*run_javascript.begin ("var t = document.getElementsByTagName('meta').filter(function(e){return e.name == 'theme-color';)[0]; t ? t.value : null;", (obj, res) => {
				var js_res = run_javascript.end (res);
			});*/

			var snap = (Cairo.ImageSurface) yield get_snapshot (WebKit.SnapshotRegion.VISIBLE,
				WebKit.SnapshotOptions.NONE, null);

			// data ist in BGRA apparently (according to testing). Docs said ARGB, but that
			// appears not to be the case
			unowned uint8[] data = snap.get_data ();
			uint8 r = data[2], g = data[1], b = data[0];

			for (var i = 4; i < snap.get_width () * 3 * 4; i += 4) {
				r = (r + data[i + 2]) / 2;
				g = (g + data[i + 1]) / 2;
				b = (b + data[i + 0]) / 2;
			}

			ui_color = "#%02x%02x%02x".printf (r, g, b);
		}

		/**
		 * @inheritDoc
		 */
		public void search (string text, bool case_insensitive, bool highlight)
		{
			var find = get_find_controller ();

			if (find.get_search_text () != null)
				find.search_finish ();

			// FIXME should we set max_match_count?
			find.search (text, 0, uint.MAX);
		}

		/**
		 * @inheritDoc
		 */
		public void next_result ()
		{
			get_find_controller ().search_next ();
		}

		/**
		 * @inheritDoc
		 */
		public void previous_result ()
		{
			get_find_controller ().search_previous ();
		}

		/**
		 * @inheritDoc
		 */
		public void open_history_item (string id)
		{
			load_uri (id);
		}

		/**
		 * @inheritDoc
		 */
		public new void reload (bool bypass_cache = false)
		{
			if (bypass_cache)
				reload_bypass_cache ();
			else
				base.reload ();
		}

		/**
		 * @inheritDoc
		 */
		public void abort_loading ()
		{
			stop_loading ();
		}

		/**
		 * @inheritDoc
		 */
		public void zoom_in ()
		{
			zoom_level = Math.fmin (5, zoom_level + 0.5);
		}

		/**
		 * @inheritDoc
		 */
		public void zoom_out ()
		{
			zoom_level = Math.fmax (0.1, zoom_level - 0.1);
		}

		/**
		 * @inheritDoc
		 */
		public bool can_zoom_in ()
		{
			// FIXME do some proper testing here
			return zoom_level < 5;
		}

		/**
		 * @inheritDoc
		 *
		 * This function won't fail if the interfaces don't match, but return false instead.
		 */
		public bool can_zoom_out ()
		{
			// FIXME do some proper testing here
			return zoom_level > 0.1;
		}

		/**
		 * @inheritDoc
		 */
		public void reset_zoom ()
		{
			zoom_level = 1.0;
		}
	}
}

