
namespace Midori
{
	public class Window : Gtk.Window
	{
		UrlBar url_bar;

		Gtk.Button back;
		Gtk.Button forward;
		Gtk.HeaderBar header;
		Granite.Widgets.DynamicNotebook notebook;

		Tab? current_tab = null;

		const string SHORTCUTS = """
			@binding-set WindowShortcuts
			{
				bind "<Control>l" { "focus-url-bar" () };
				bind "<Alt>Left" { "go-back" () };
			}

			.midori-window
			{
				gtk-key-bindings: WindowShortcuts;
			}
		""";

		[Signal (action = true)]
		public virtual signal void focus_url_bar () { url_bar.grab_focus (); }

		[Signal (action = true)]
		public virtual signal void go_back () { current_tab.back (); }

		public Window ()
		{
			try {
				var shortcuts_css = new Gtk.CssProvider ();
				shortcuts_css.load_from_data (SHORTCUTS, -1);

				var style_context = get_style_context ();
				style_context.add_class ("midori-window");
				style_context.add_provider (shortcuts_css, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
			} catch (Error e) {
				warning (e.message);
			}

			header = new Gtk.HeaderBar ();
			header.show_close_button = true;

			back = new Gtk.Button.from_icon_name ("go-previous-symbolic", Gtk.IconSize.LARGE_TOOLBAR);
			forward = new Gtk.Button.from_icon_name ("go-next-symbolic", Gtk.IconSize.LARGE_TOOLBAR);
			var menu = new Gtk.Button.from_icon_name ("document-properties", Gtk.IconSize.LARGE_TOOLBAR);

			menu.clicked.connect (() => {
				if (current_tab == null)
					return;

				current_tab.go_to_native ("downloads-view", new DownloadsView ());
			});

			back.clicked.connect (() => {
				current_tab.back ();
			});

			forward.clicked.connect (() => {
				current_tab.forward ();
			});

			header.pack_start (back);
			header.pack_start (forward);
			header.pack_end (menu);

			url_bar = new UrlBar ();
			url_bar.hexpand = true;
			// FIXME no constant width
			url_bar.width_request = 500;
			url_bar.submitted.connect ((text) => {
				var uri = Utils.magic_uri (text, true, false);

				// no uri detected, must be a search query
				if (uri == null) {
					try {
						var manager = (SearchEngineManager) App.get_default ().get_feature ("search-engines");
						// check for tokens for a certain engine
						uri = manager.query_matches_token (text);

						if (uri == null) {
							// TODO connect to settings to decide which to use
							uri = URI.for_search (manager.search_engines.nth_data (0).uri, text);
						} else
							uri = URI.for_search (uri, text.substring (text.index_of (" ")));
					} catch (Error e) {}
				}

				current_tab.go_to_uri (uri);
				current_tab.grab_focus ();
			});
			url_bar.load_action.connect (() => {
				if (!current_tab.implements (typeof (Loadable)))
					return;

				if (current_tab.loading) {
					current_tab.abort_loading ();
				} else {
					current_tab.reload ();
				}
			});
			header.custom_title = url_bar;
			header.get_style_context ().add_class ("midori-titlebar");

			set_titlebar (header);

			notebook = new Granite.Widgets.DynamicNotebook ();
			notebook.show_icons = true;
			notebook.tab_switched.connect ((old_tab, new_tab) => {
				update_current_tab ((Tab) new_tab.page);
			});
			notebook.new_tab_requested.connect (() => {
				add_tab (null, true);
			});
			notebook.tab_removed.connect (() => {
				if (notebook.n_tabs < 1)
					quit ();
			});

			add (notebook);
			add_tab ("http://google.com");

			tab_loading_changed (null, null);
			navigation_items_changed (null, null);

			// apply_theme_color ("#f04530");
		}

		void update_current_tab (Tab new_tab)
		{
			if (current_tab != null) {
				current_tab.notify["loading"].disconnect (tab_loading_changed);
				current_tab.notify["can-go-back"].disconnect (navigation_items_changed);
				current_tab.notify["can-go-forward"].disconnect (navigation_items_changed);
				current_tab.notify["security-status"].disconnect (security_status_changed);
				current_tab.notify["uri"].disconnect (uri_changed);
				current_tab.notify["ui-color"].disconnect (ui_color_changed);
			}

			current_tab = new_tab;
			current_tab.notify["loading"].connect (tab_loading_changed);
			current_tab.notify["can-go-back"].connect (navigation_items_changed);
			current_tab.notify["can-go-forward"].connect (navigation_items_changed);
			current_tab.notify["security-status"].connect (security_status_changed);
			current_tab.notify["uri"].connect (uri_changed);
			current_tab.notify["ui-color"].connect (ui_color_changed);

			tab_loading_changed ();
			navigation_items_changed ();
			security_status_changed ();
			uri_changed ();
			ui_color_changed ();
		}

		void tab_loading_changed (Object? tab = null, ParamSpec? pspec = null)
		{
			url_bar.page_is_loading = ((Tab) current_tab).loading;
		}

		void security_status_changed (Object? tab = null, ParamSpec? pspec = null)
		{
			url_bar.page_security_status = current_tab.security_status;
		}

		void uri_changed (Object? tab = null, ParamSpec? pspec = null)
		{
			url_bar.page_uri = current_tab.uri;
		}

		void navigation_items_changed (Object? tab = null, ParamSpec? pspec = null)
		{
			back.sensitive = current_tab.can_go_back;
			forward.sensitive = current_tab.can_go_forward;
		}

		void ui_color_changed (Object? tab = null, ParamSpec? pspec = null)
		{
			apply_theme_color (current_tab.ui_color);
		}

		Gtk.CssProvider? header_custom_css = null;

		/**
		 * Set the headerbar's color to the given color. If color is null, reset
		 * to the theme's color.
		 *
		 * @param color The color to switch to
		 */
		void apply_theme_color (string? color)
		{
			var style_context = header.get_style_context ();

			if (header_custom_css != null)
				style_context.remove_provider (header_custom_css);

			if (color == null)
				return;

			header_custom_css = new Gtk.CssProvider ();
			header_custom_css.load_from_data ("@define-color colorPrimary %s;"
				.printf (color), -1);

			style_context.add_provider (header_custom_css, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION);
		}

		public Tab add_tab (string? uri = null, bool focus = false)
		{
			var tab = new Tab ();
			var label = new Granite.Widgets.Tab (tab.page_title, tab.icon, tab);

			tab.notify["loading"].connect (() => {
				label.working = tab.loading;
			});
			tab.notify["page-title"].connect (() => {
				label.label = tab.page_title;
			});
			tab.notify["icon"].connect (() => {
				label.icon = tab.icon;
			});

			if (uri != null)
				tab.go_to_uri (uri);

			notebook.insert_tab (label, -1);

			if (focus)
				notebook.current = label;

			return tab;
		}

		public void quit ()
		{
			destroy ();
		}
	}
}

